#--------------------------------------------------------------------------
#
#   ALHM - Level
#
#--------------------------------------------------------------------------

from __future__ import division
import os
from collections import defaultdict
from random import random, normalvariate, choice
from itertools import izip
from xml.dom import minidom
import pygame
from pygame import draw, Color, Rect
from pygame.transform import rotate
from albow.resource import get_image, get_font, get_sound, _get_image
from albow.image_array import get_image_array
from albow.dialogs import alert
from dialogs import ImageDialog

from constants import *
from functions import *
from swarm import Swarm
from player import Player

sting_rate = 5
sting_cooldown_time = 0.4

debug_draw_walls = False
# debug_draw_walls = True

#--------------------------------------------------------------------------

def bounding_rect(points):
    xs = [p[0] for p in points]
    ys = [p[1] for p in points]
    x0 = min(xs)
    y0 = min(ys)
    x1 = max(xs)
    y1 = max(ys)
    return Rect(x0, y0, x1 - x0, y1 - y0)

#--------------------------------------------------------------------------

class Cell(object):

    def __init__(self):
        self.distance_from_goal = 1000
        self.connected = [True, True, True, True]
    
    def connected_directions(self):
        connected = self.connected
        for dir in directions:
            if connected[dir]:
                yield dir

#--------------------------------------------------------------------------

class Gadget(object):

    def switched(self, level):
        pass

#--------------------------------------------------------------------------

def order_points(p1, p2):
    if p1[0] > p2[0] or p1[1] > p2[1]:
        p1, p2 = p2, p1
    return (p1, p2)

class Wall(Gadget):

    def __init__(self, points):
        self.points = order_points(*points)
        self.orientation = HORIZONTAL if self.points[0][1] == self.points[1][1] else VERTICAL
    
    def __iter__(self):
        return iter(self.points)

    def is_solid(self):
        return True

    def draw(self, surf):
        if debug_draw_walls:
            self.draw_with_color(surf, black)
    
    def draw_with_color(self, surf, c):
        p1, p2 = self
        draw.line(surf, c, p1, p2)

#--------------------------------------------------------------------------

class Door(Wall):

    h_images = [
        get_image("door-h-bg.png"),
        get_image("door-h-open.png"),
        get_image("door-h-closed.png"),
    ]
    v_images = [rotate(img, 90) for img in h_images]
    images = [h_images, v_images]
    
    sounds = [
        get_sound("door-open.ogg"),
        get_sound("door-close.ogg"),
    ]

    def __init__(self, points, tag):
        Wall.__init__(self, points)
        self.tag = tag
        self.is_closed = True
    
    def is_solid(self):
        return self.is_closed
        
    def draw(self, surf):
        #self.draw_with_color(surf, red if self.is_closed else green)
        images = self.images[self.orientation]
        x, y = self.points[0]
        p = (x - 16, y - 16)
        surf.blit(images[0], p)
        surf.blit(images[self.is_closed + 1], p)

    def switched(self, level):
        self.is_closed = not self.is_closed
        level.update_cell_walls(self)
        self.sounds[self.is_closed].play()

#--------------------------------------------------------------------------

class Switch(Gadget):

    images = [
        get_image("lever-off.png"),
        get_image("lever-on.png"),
    ]

    def __init__(self, points, tag):
        self.bounds = bounding_rect(points)
        self.tag = tag
        self.is_on = False
    
    def draw(self, surf):
        #draw.rect(surf, black, self.bounds, 1)
        img = self.images[self.is_on]
        r = img.get_rect()
        r.center = self.bounds.center
        surf.blit(img, r)
    
    def activate(self, level):
        #print "Switch.activate:", self.tag
        self.is_on = not self.is_on
        for obj in level.tags[self.tag]:
            obj.switched(level)
        level.find_paths()

#--------------------------------------------------------------------------

class Trap(Gadget):

    def trample(self, level):
        pass

    def begin_frame(self, dt):
        pass

#--------------------------------------------------------------------------

class Cobweb(Trap):

    intact = get_image("cobweb.png")
    broken = get_image("cobweb-broken.png")
    catch_rate = 2.0

    def __init__(self, points):
        self.bounds = bounding_rect(points)
        self.is_active = True
    
    def draw(self, surf):
        #draw.rect(surf, red, self.bounds, 0 if self.is_active else 1)
        img = self.intact if self.is_active else self.broken
        r = img.get_rect()
        r.center = self.bounds.center
        surf.blit(img, r)
    
    def trample(self, level):
        self.is_active = False
    
    def is_operational(self):
        return self.is_active

#--------------------------------------------------------------------------

class Sprayer(Trap):

    spray = get_image("spray-cloud.png")
    sound = get_sound("spray.ogg")
    catch_rate = 2.0
    spray_time = 20.0
    
    def __init__(self, points):
        self.bounds = bounding_rect(points)
        self.time_left = self.spray_time
        self.is_active = False
    
    def draw(self, surf):
        if self.is_operational():
            img = self.spray
            r = img.get_rect()
            r.center = self.bounds.center
            surf.blit(img, r)
    
    def begin_frame(self, dt):
        if self.is_operational():
            self.time_left -= dt
    
    def switched(self, level):
        self.is_active = not self.is_active
        if self.is_operational():
            self.sound.play()
        else:
            self.sound.stop()

    def is_operational(self):
        return self.is_active and self.time_left > 0

#--------------------------------------------------------------------------

class Level(object):

    cell_size = 128
    cells_wide = 8
    cells_high = 6
    #goal_cc = (3, 2)
    #swarm_start_cc = (0, 0)

#     walls = [
#         ((20, 20), (100, 20)),
#         ((100, 40), (100, 80)),
#     ]

    def level_is_completed(self):
        return self.game_outcome
    
#     def begin_frame(self):
#         pass
    
    def read_from_file(self, lvl_path):
        def number(s):
            #print "number:", s
            return float(s)
        def parse_point(s):
            return map(number, s[1:].split(","))
        base = os.path.splitext(lvl_path)[0]
        svg_path = base + ".svg"
        self.bg_path = base + ".png"
        doc = minidom.parse(svg_path)
        walls = []
        switches = []
        traps = []
        tags = defaultdict(list)
#         path_strings = [path.getAttribute('d') for path in doc.getElementsByTagName('path')]
#         for line_str in path_strings:
        for path in doc.getElementsByTagName('path'):
            points_str = path.getAttribute('d')
            #print "points_str:", points_str
            points = []
            for s in points_str.split():
                if s[:1] in "ML":
                    points.append(parse_point(s))
            color = path.getAttribute('stroke')
            width = float(path.getAttribute('stroke-width'))
            n = len(points)
            if n == 3:
                #print "Switch", color, points
                switches.append(Switch(points, color))
            elif n == 4:
                trap = Sprayer(points)
                traps.append(trap)
                tags[color].append(trap)
                #print "Sprayer", color, points
            elif n == 5:
                cc = self.cc_containing(points[0])
                if color == "#f00":
                    self.swarm_start_cc = cc
                elif color == "#0f0":
                    self.goal_cc = cc
            elif n == 6:
                #print "Trap", points
                traps.append(Cobweb(points))
            else:
                if color == "#000":
                    wall = Wall(points)
                else:
                    wall = Door(points, color)
                    wall.is_closed = width > 6
                    tags[color].append(wall)
                walls.append(wall)
        #for doorlist in tags.itervalues():
        #   doorlist[0].is_closed = not doorlist[0].is_closed
        self.walls = walls
        self.switches = switches
        self.traps = traps
        self.tags = tags
        
    def draw(self, surf):
        surf.blit(self.background_image, (0, 0))
        #self.debug_draw_goal_distances(surf)
        for wall in self.walls:
            wall.draw(surf)
        for switch in self.switches:
            switch.draw(surf)
        for trap in self.traps:
            trap.draw(surf)
        self.player.draw(surf)
        self.swarm.draw(surf)

    def debug_draw_goal_distances(self, surf):
        for cy in xrange(self.cells_high):
            for cx in xrange(self.cells_wide):
                cc = (cx, cy)
                cell = self.get_cell(cc)
                text = font.render(str(cell.distance_from_goal), True, black)
                surf.blit(text, self.cell_centre(cc))

    def prepare(self):
        self.background_image = _get_image(self.bg_path)
        self.cells = [[Cell() for col in xrange(self.cells_wide)] for row in xrange(self.cells_high)]
        self.wall_off_cells()
        self.find_paths()
        self.player = Player(self, self.cell_centre(self.goal_cc))
        self.swarm = Swarm(self.cell_centre(self.swarm_start_cc))
        self.game_outcome = 0
        self.sting_cooldown = 0
        self.goal_inaccessible = False

    def begin_frame(self, dt):
        self.swarm.begin_frame(self, dt)
        self.player.begin_frame(self, dt)
        self.check_for_stings(dt)
        for trap in self.traps:
            trap.begin_frame(dt)
        if self.swarm.size == 0:
            self.game_outcome = 1
    
    def check_for_stings(self, dt):
        if self.sting_cooldown > 0:
            self.sting_cooldown -= dt
        else:
            pp = self.player.position
            for hp in self.swarm.positions:
                if points_near(hp, pp, 25):
                    if chance(dt * sting_rate):
                        self.player.sting(self)
                        sting_sound.play()
                        self.sting_cooldown = sting_cooldown_time
                        return

    def find_paths(self):
        def pf(cc, dist):
            cell = self.get_cell(cc)
            if cell:
                if cell.distance_from_goal > dist:
#                     print "find_paths:", cc, "=", dist
                    cell.distance_from_goal = dist
                    for dir in directions:
                        if cell.connected[dir]:
                            pf(add(cc, dirvec[dir]), dist + 1)
        for row in self.cells:
            for cell in row:
                cell.distance_from_goal = 1000
        pf(self.goal_cc, 0)
        self.check_goal_accessible()

    def check_goal_accessible(self):
        cell = self.get_cell(self.swarm_start_cc)
        if cell.distance_from_goal < 1000:
            self.goal_inaccessible = False
        elif not self.goal_inaccessible:
            self.goal_inaccessible = True
            #alert("Conservatory blocked")
            self.report_conservatory_blocked()
    
    def report_conservatory_blocked(self):
        dlog = ImageDialog("conservatory-blocked.png", "buttons.png",
            responses = [(1, "sorry.png")])
        dlog.present()

    def cell_centre(self, cc):
        cx, cy = cc
        cs = self.cell_size
        x = (cx + 0.5) * cs
        y = (cy + 0.5) * cs
        return (x, y)

    def get_cell(self, cc):
        cx, cy = cc
        if 0 <= cy < self.cells_high and 0 <= cx < self.cells_wide:
            return self.cells[cy][cx]

    def set_cell_connected(self, cc, dir, state):
        cell = self.get_cell(cc)
        if cell:
            cell.connected[dir] = state

    def cc_containing(self, p):
        x, y = p
        cs = self.cell_size
        cx = int(round(x)) // cs
        cy = int(round(y)) // cs
        return (cx, cy)

    def wall_off_cells(self):
        for wall in self.walls:
            self.update_cell_walls(wall)
    
    def update_cell_walls(self, wall):
        conn = not wall.is_solid()
        p0, p1 = wall
        cc0 = self.cc_containing(p0)
        cc1 = self.cc_containing(p1)
        if cc0[1] == cc1[1]:
            cy = cc0[1]
            for cx in xrange(cc0[0], cc1[0]):
                self.set_cell_connected((cx, cy - 1), DOWN, conn)
                self.set_cell_connected((cx, cy), UP, conn)
        else:
            cx = cc0[0]
            for cy in xrange(cc0[1], cc1[1]):
                self.set_cell_connected((cx - 1, cy), RIGHT, conn)
                self.set_cell_connected((cx, cy), LEFT, conn)

    def start_moving(self, dir):
        self.player.start_moving(dir)

    def stop_moving(self, dir):
        self.player.stop_moving(dir)

    def activation_cmd(self):
        r = self.player.activation_rect()
        for switch in self.switches:
            if r.colliderect(switch.bounds):
                switch.activate(self)
